// Created by Josh Heyde
// Last updated 27/01/09

#include "C_AI.h"
#include "C_Wave.h"
#include "C_Route.h"
#include "C_WorldEntity.h"
#include "C_World.h"
#include <assert.h>
#include <sstream>
#include <ctime>

using namespace std;

// constructor
C_AI::C_AI(const int &numOfRoutes, const int &fBE, const int &lBE, const int &fAE, const int &lAE)
: NUM_OF_ROUTES(numOfRoutes), firstBasicEnemy(fBE), lastBasicEnemy(lBE), firstAdvancedEnemy(fAE), lastAdvancedEnemy(lAE),
	currentWave(0), previousTime(0), DELAY(1000)
{
	// seeds the random number generator
	srand(time(0));
}

// destructor
C_AI::~C_AI()
{
	// deletes all created waves
	for (std::vector<C_Wave*>::const_iterator iter = v_waveVector.begin(); iter < v_waveVector.end(); iter++)
		delete *iter;

	// deletes all created routes
	for (std::vector<C_Route*>::const_iterator iter = v_routeVector.begin(); iter < v_routeVector.end(); iter++)
		delete *iter;
}

// other member functions

template <class T>
std::string toString(const T & t)
{
    std::ostringstream oss; // create a stream
    oss << t;              // insert value to stream
    return oss.str();      // return as a string
}

// initialises the component
bool C_AI::intialise(std::vector <C_WorldEntity*> v_entityVector)
{
	// creates the various routes for the enemies to follow
	for(int i = 0; i < NUM_OF_ROUTES; i++)
	{
		// creates a new route and pushes onto the route vector
		v_routeVector.push_back(new C_Route("AI\\route" + toString(i + 1) + ".txt"));
		
		// ensures the route can correctly intialise
		if(!v_routeVector[i]->initialise())
		{
			WORLD.userMessage("Failed to create route " + toString((i + 1)), "Error");
			return false;
		}
	}

	// creates the various waves of enemies to attack the player
	// assigning them random routes to somewhat prevent predictability
	v_waveVector.push_back(new C_Wave(rand() % NUM_OF_ROUTES, firstBasicEnemy, firstBasicEnemy + 1, firstBasicEnemy + 2, firstBasicEnemy + 3, firstBasicEnemy + 4));
	v_waveVector.push_back(new C_Wave(rand() % NUM_OF_ROUTES, firstBasicEnemy, firstBasicEnemy + 1, firstBasicEnemy + 2, firstBasicEnemy + 3, firstAdvancedEnemy));
	v_waveVector.push_back(new C_Wave(rand() % NUM_OF_ROUTES, firstBasicEnemy, firstBasicEnemy + 1, firstBasicEnemy + 2, firstAdvancedEnemy, firstAdvancedEnemy + 1));
	v_waveVector.push_back(new C_Wave(rand() % NUM_OF_ROUTES, firstAdvancedEnemy, firstBasicEnemy, firstAdvancedEnemy + 1, firstBasicEnemy + 1, firstAdvancedEnemy + 2));
	v_waveVector.push_back(new C_Wave(rand() % NUM_OF_ROUTES, firstAdvancedEnemy, firstAdvancedEnemy + 1, firstAdvancedEnemy + 2, firstAdvancedEnemy + 3, firstAdvancedEnemy + 4));

	// intialises and spawns the first wave
	v_waveVector[currentWave]->intialise(v_entityVector, v_routeVector[v_waveVector[currentWave]->getRouteReference()]);

	// signals intialisation has been successful
	return true;
}

// updates the AI component
void C_AI::update(std::vector<C_WorldEntity*> v_entityVector, const float &currentTime)
{
	// spawns an enemy if there are any left to spawn and enough time has passed
	if(currentTime > previousTime + DELAY)
	{
		// checks to see if all enemies have been spawned		
		if(v_waveVector[currentWave]->allEnemiesSpawned())
		{
			// moves onto the next wave if the current one has been completed
			if(v_waveVector[currentWave]->isWaveComplete(v_entityVector))
			{
				// checks to see if any waves are remaining
				if(currentWave < v_waveVector.size() - 1)
				{
					currentWave++;
					v_waveVector[currentWave]->intialise(v_entityVector, v_routeVector[v_waveVector[currentWave]->getRouteReference()]);
				}
				else
				{
					// if none are remaining, all waves are reset
					currentWave = 0;
					
					for(size_t i = 0; i < v_waveVector.size(); i++)
					{
						v_waveVector[i]->resetWave(v_entityVector);
						// re assigns the route of the enemies to decrease predictability
						v_waveVector[i]->setRouteReference(rand() % NUM_OF_ROUTES);
					}

					v_waveVector[currentWave]->intialise(v_entityVector, v_routeVector[v_waveVector[currentWave]->getRouteReference()]);
				}
			}
		}
		else
		{
			v_waveVector[currentWave]->spawnEnemy(v_entityVector);
		}

		previousTime = currentTime;
	}

	// updates the current wave, if there are any remaining
	v_waveVector[currentWave]->update(v_entityVector, v_routeVector[v_waveVector[currentWave]->getRouteReference()]);
}

// initialisation of the singleton
C_AI* C_AI::ai = NULL;

// creation of a singleton
void C_AI::createAI(const int &numOfRoutes, const int &fBE, const int &lBE, const int &fAE, const int &lAE)
{
	if(!ai)
		ai = new C_AI(numOfRoutes, fBE, lBE, fAE, lAE);
}

// deletion of a singleton
void C_AI::deleteAI()
{
	// error handling statements
	// asserting certain statements for this function
	assert(ai != NULL);

	delete ai;
}